/* Mueller Matthias simmmuel Informatik */
/* wsort - program to check out qsort() function */
/* Exercise 2 */

/* includes */
#include <stdio.h>  /* fgets */
#include <malloc.h> /* malloc */
#include <string.h> /* strcmp */
#include <stdlib.h> /* qsort */
#include <search.h> /* qsort */
#include <errno.h>  /* errno */

/* prototypes */
void arg_print(const char *message, int argc, char **argv);
long strings_read(int *string_number, char ***string_array_passed);
void strings_sort(const int string_number, char *string_array[]);
int strings_sort_compare(const void *string1, const void *string2);

int main(int argc, char *argv[]) /* Windows: , char *envp[]) */
{
    int string_number=0;
    char **string_array=NULL; /* pointer on an array of pointers */

    errno=0; /* do not re-use eventually existing previous error number */

    /* arg_print("Program arguments:", argc, argv); */

    if (argc==1) /* merely program path passed as argument */
    {
        arg_print ("Please enter words that are to be sorted:", 0, NULL);

        if (strings_read(&string_number, &string_array)==0) /* verify */
        {
            perror(" Fehler beim Einlesen"); /* e.g. ' Fehler beim Einlesen: No error */
            exit (EXIT_FAILURE);
        }
    }
    else
    {
        /* make 'input strings' point to argument strings */
        string_number=argc;
        string_array=argv;

        /* remove program path */
        string_number--;
        string_array++;
    }

    strings_sort(string_number, string_array); /* sort any kind of strings (inputted words or program arguments) */
    arg_print("Words sorted:", string_number, string_array);

    return 0;
}

void arg_print(const char *message, int argc, char *argv[])
{
    /* prints passed program arguments to standard-out; message will be printed in an extra line */

    int argv_length=0;
    int i=0;

    printf ("%s", message);
    printf ("\n");

    for (i=0; i<argc; i++)
    {
        /* print string */
        printf ("%s", argv[i]);

        /* begin next line if necessary */
        argv_length=strlen(argv[i]);
        if (argv_length>0) /* avoid access violation */
        {
            if ((*(argv[i]+argv_length-1))!='\n')
                printf ("\n");
        }
    }
}

long strings_read(int *string_number, char ***string_array_passed)
{
    /* reads words from standard-in; returns non-zero (1) for success or zero for error */
    char *string_temp=NULL; /* IMPORTANT: set to NULL or there will be a mysterious error */
    char *string_temp_original=NULL;
    char **string_array=NULL;

    string_array=*string_array_passed;

    /* NOTE: we pass string_array with 'three asterisks': */
    /* Asterisk 1: to create a field of POINTERS */
    /* Asterisk 2: to have a pointer on that field (the data format passed by OS in arguments) */
    /* Asterisk 3: to pass field by reference */

    while (1)
    {
        string_temp=(char *)malloc((size_t)(102*sizeof(char)));
        string_temp_original=string_temp;

        if (string_temp==NULL) /* verify */
        {
            return 0;
        }

        string_temp=fgets(string_temp, 102, stdin);

        /* check for 'special' input */
        if (string_temp==NULL) /* error or EOF */
        {
            free(string_temp_original); /* important */
            return 1;
        }

        /* process 'normal' input */
        if (strlen(string_temp)==1)
        {
            if (string_temp[0]=='\n') /* empty line */
            {
                free(string_temp); /* will be re-allocated */
                string_temp=NULL;
                string_temp_original=string_temp;
                /* return 1; enable for Windows systems (Ctrl-D Return failed) */
            }
        }
        else
        {
            if (strlen(string_temp)==101) /* more than 100 chars entered */
            {
                fprintf (stderr, " Please enter words with a length up to 100 chars only !\n (Last word was ignored)\n");

                /* NOTE: the remaining chars (of pos 102+) would be instantly written to temp_string */
                /* in next input cycle. Therefore we throw away the whole input buffer content. */
                /* MSDN says: "If the stream is open for input, fflush clears the contents of the buffer." */

                while (getc(stdin)!='\n'); /* throw away rest of input buffer */

                free(string_temp);
                string_temp=NULL;
                string_temp_original=string_temp;
            }
            else /* ok, process input */
            {
                (*string_number)++;

                string_array=(char **)realloc((void *)string_array, (size_t)(sizeof(char **)*(*string_number)));

                if (string_array==NULL) /* verify */
                {
                    free(string_temp);
                    free(string_array);
                    (*string_number)=0;
                    return 0;
                }
                string_array[(*string_number)-1]=string_temp;
                *string_array_passed=string_array;
                string_temp=NULL; /* we still have a pointer to memory in string_array */
                string_temp_original=string_temp;
            }
        }
    }
}

void strings_sort(const int string_number, char *string_array[])
{
    /* sorts strings using a pre-defined quick sort algorithm */
    qsort((void *)string_array, (size_t)string_number, sizeof(char *), strings_sort_compare);
}

int strings_sort_compare(const void *string1, const void *string2)
{
    /* compare function used by ANSI-library function qsort */
    return strcmp(*((char **)string1), *((char **)string2));
}
