/* Mller Matthias simmmuel Informatik 23.11.03 */
/* printdir - program that behaves similar to UNIX-commando ls(1) */
/* (at least it should do so) */

/* includes */
#include <stdio.h>              /* printf */
#include <errno.h>              /* errno */
#include <sys/types.h>          /* opendir */
#include <dirent.h>             /* opendir */
#include <sys/stat.h>           /* stat */
#include <unistd.h>             /* readlink */
#include <malloc.h>             /* malloc */
#include <string.h>             /* strncpy */
#include <grp.h>                /* getgrgid */
#include <pwd.h>                /* getpwuid */
#include <bits/posix1_lim.h>    /* PATH_MAX */

/* defines */
#define PRIVATE_PATH_MAX    PATH_MAX+1 /* use like PATH_MAX (one additionally char for null terminator) */

/* prototypes */
int main(int, char **);
int printdir(const char *); /* main function that displays directory content */
int help_isfilehidden(char *); /* helper function */
void help_strconcat(char *dest, const char *src1, const char *src2, size_t string_length); /* helper function */

int main(int argc, char *argv[])
{
    char *printdir_dir=NULL; /* list directory (slash-terminated) */
    int printdir_dir_length=0;
    int i=0;

    /* verify */
    if (argc<=1)
    {
        fprintf(stderr, "Please pass name of directory that is to be listed as argument !\n");
        return 0;
    }

    /* preset */
    printdir_dir=(char *)malloc(PRIVATE_PATH_MAX-1);
    if (printdir_dir==NULL)
        return 0;

    /* begin */
    for(i=1; i<argc; i++) /* first argument is program path, ignore it */
    {
        /* initialize printdir_dir with argument */
        help_strconcat(printdir_dir, argv[i], "", PRIVATE_PATH_MAX);

        /* verify printdir_dir is slash-terminated */
        printdir_dir_length=strlen(printdir_dir);
        if (printdir_dir_length>0)
            if ((*(&(*printdir_dir)+printdir_dir_length-1))!='/')
            {
                help_strconcat(printdir_dir, argv[i], "/", PRIVATE_PATH_MAX);
            }

        /* read and display directory data */
        if (printdir(printdir_dir)==0)
            perror("Error reading directory and/or file data in printdir()");
    }

    return 0;
}

int printdir(const char *dir)
{
    /* prints content of dir, returns non-zero (1) for success or zero for error */
    DIR *dir_var=NULL;
    struct dirent *dirent_var=NULL;
    char *dir_and_file=NULL;
    struct stat *stat_var=NULL;
    char *file_name=NULL;
    char *link_string=NULL;
    char *rights_string=NULL;
    struct passwd *passwd_var=NULL;
    struct group *group_var=NULL;

    /* preset */
    errno=0; /* important, we need to check if there's really an error when readdir() return NULL */

    dir_and_file=(char *)malloc((size_t)PRIVATE_PATH_MAX);
    if (dir_and_file==NULL)
        return 0;

    rights_string=(char *)malloc((size_t)11);
    if (rights_string==NULL)
        return 0;
    else
        rights_string[10]='\0'; /* string terminator */

    file_name=(char *)malloc((size_t)PRIVATE_PATH_MAX);
    if (file_name==NULL)
        return 0;
    else
        file_name[PRIVATE_PATH_MAX-1]='\0';

    /* begin */
    dir_var=opendir(dir); /* get directory handle */
    if (dir_var==NULL) /* verify directory can be read */
        return 0;

    printf("%s\n", dir); /* print this directory's name */

    dirent_var=(struct dirent *)malloc((size_t)(sizeof(struct dirent)));
    if (dirent_var==NULL)
        return 0;

    while (1)
    {
        dirent_var=readdir(dir_var); /* get next directory entry */
        if (dirent_var==NULL) /* either error or finished */
        {
            if (errno==0)
                return 1;
            else
                return 0;
        }

        /* if we have a hidden file then do not display */
        if (help_isfilehidden(dirent_var->d_name)==0)
        {
            help_strconcat(dir_and_file, dir, (char *)dirent_var->d_name, PRIVATE_PATH_MAX-1);

            stat_var=(struct stat *)malloc((size_t)(sizeof(struct stat)));
            if (stat_var==NULL)
                return 0;

            if (lstat(dir_and_file, stat_var)==(-1))
                return 0;

            /* printf("d_type: %d\n", (int)dirent_var->d_type); */ /* always 0 ?!? */
            /* HINWEIS: Verzeichnis-Namen werden auch ausgegeben, aber es hat in der */
            /* Aufgabenstellung niemand gesagt, dass wir Verzeichnisse NICHT anzeigen sollen !!! */

            link_string=(char *)malloc(PRIVATE_PATH_MAX);
            if (link_string==NULL)
                return 0;
            else
                link_string[PRIVATE_PATH_MAX-1]='\0';

            /* get file name to display */

            if (S_ISLNK(stat_var->st_mode)) /* do NOT just check if true */
            {
                strncpy(file_name, (char *)dirent_var->d_name, PRIVATE_PATH_MAX-1);

                if (readlink(dir_and_file, link_string, PRIVATE_PATH_MAX-1)==(-1))
                    return 0;

                strncat(file_name, " -> ", PRIVATE_PATH_MAX-1);
                strncat(file_name, link_string, PRIVATE_PATH_MAX-1);
            }
            else
            {
                strncpy(file_name, (char *)dirent_var->d_name, PRIVATE_PATH_MAX-1);
            }

            /* create rights-string */

            if(S_ISREG(stat_var->st_mode))
                rights_string[0]='-';
            else if(S_ISLNK(stat_var->st_mode))
                rights_string[0]='l';
            else if(S_ISDIR(stat_var->st_mode))
                rights_string[0]='d';
            else if(S_ISCHR(stat_var->st_mode))
                rights_string[0]='c';
            else if(S_ISBLK(stat_var->st_mode))
                rights_string[0]='b';
            else if(S_ISFIFO(stat_var->st_mode))
                rights_string[0]='f';
            /* doesn't work
            else if(S_ISSOCK(stat_var->st_mode))
                rights_string[0]='s'; */

            if (stat_var->st_mode & S_IRUSR) rights_string[1]='r';
            else rights_string[1]='-';
            if (stat_var->st_mode & S_IWUSR) rights_string[2]='w';
            else rights_string[2]='-';
            if (stat_var->st_mode & S_IXUSR) rights_string[3]='x';
            else rights_string[3]='-';
            if (stat_var->st_mode & S_IRGRP) rights_string[4]='r';
            else rights_string[4]='-';
            if (stat_var->st_mode & S_IWGRP) rights_string[5]='w';
            else rights_string[5]='-';
            if (stat_var->st_mode & S_IXGRP) rights_string[6]='x';
            else rights_string[6]='-';
            if (stat_var->st_mode & S_IROTH) rights_string[7]='r';
            else rights_string[7]='-';
            if (stat_var->st_mode & S_IWOTH) rights_string[8]='w';
            else rights_string[8]='-';
            if (stat_var->st_mode & S_IXOTH) rights_string[9]='x';
            else rights_string[9]='-';

            /* get user and group name (ID) */

            passwd_var=getpwuid(stat_var->st_uid);
            group_var=getgrgid(stat_var->st_gid);

            /* print out whole stuff */

            printf("%s\t",
                   rights_string);

            if (passwd_var==NULL) /* display numeric value (UID) if not resolvable */
                printf("%d\t", stat_var->st_uid);
            else
                printf("%s\t", passwd_var->pw_gecos);

            if (group_var==NULL) /* display numeric value (UID) if not resolvable */
                printf("%d\t", stat_var->st_gid);
            else
                printf("%s\t", group_var->gr_name);

            printf("%d\t%d\t%s\n",
                   (int)stat_var->st_nlink,
                   (int)stat_var->st_size,
                   file_name);
        } /* do not display hidden file */
    }
    return 1; /* maybe nothing displayed as all hidden or directory, no error */
}

void help_strconcat(char *dest, const char *src1, const char *src2, size_t string_length)
{
    /* puts src1 and src2 one after another into dest */
    memset(dest, 0, string_length);
    strncpy(dest, src1, string_length-1); /* retain string terminator */
    strncat(dest, (char *)src2, string_length-1); /* retain string_terminator */
}

int help_isfilehidden(char *file)
{
    /* returns 1 if file begins with a point and is to be hidden, zero otherwise */
    if (strlen(file)>0)
    {
        if (file[0]=='.')
            return 1;
        else
            return 0;
    }
    else
        return 0; /* doesn't begin with a point and thus is not to be hidden */
}
