/* Mller Matthias simmmuel Diplominformatik */
/* Lsung Aufgabe 4 - mysh */
/* 30.11.03 & 01/02.12.03 */

/* includes */
#include <errno.h>      /* perror */
#include <stdlib.h>     /* malloc */
#include <stdio.h>      /* fgets */
#include <unistd.h>     /* sysconf */
#include <sys/types.h>  /* fork */
#include <sys/wait.h>   /* wait */
#include <string.h>     /* strlen */

/* prototypes */
int main();
int mysh_getline(char **);
int mysh_processline(char *);
int line_to_arg(const char *line, int *argc, char ***argv);
void line_replace_newline(char **);
void arg_free(int *argc, char ***argv);

int main()
{
    char *line=NULL;

    errno=0;

    while(1)
    {
        switch (mysh_getline(&line))
        {
        case 0:
        {
            perror("Error in mysh_getline()");
            if (line!=NULL) free(line); /* maybe error in malloc() */
            line=NULL;
            exit(-1);
        }
        case 2:
        {
            printf("\n");
            if (line!=NULL) free(line); /* maybe reset by fgets() */
            line=NULL;
            return 0;
        }
        }
        switch (mysh_processline(line))
        {
        case 0:
        {
            perror("Error in mysh_processline()");
            free(line);
            line=NULL;
            exit(-1);
            break;
        }
        case 3:
        {
            free(line);
            line=NULL;
            exit(-1);
            break;
        }
        }

        free(line); /* free before next use */
        line=NULL;
    }

    return 0;
}

int mysh_getline(char **line)
{
    /* returns 1 for success or 0 for error or 2 for user abort */

    char *line_original=NULL; /* used for freeing only */
    int line_length_max=0; /* max. length of command line */

    /* reset */
    errno=0; /* reset in any case as used to check for new errors */

    /* begin */
    line_length_max=sysconf(_SC_ARG_MAX);
    if (line_length_max==(-1))
        return 0;

    /* initialize line */
    *line=(char *)malloc(line_length_max*sizeof(char));
    if ((*line)==NULL)
        return 0;

    /* read line */
    printf("mysh> ");
    line_original=*line;
    *line=fgets(*line, line_length_max, stdin);

    if ((*line)==NULL) /* seg. fault if using fgets() here */
    {
        if (line_original!=NULL) free(line_original); /* important, or leaking */
        if (errno==0)
            return 2;
        else
            return 0;
    }

    return 1;
}

int mysh_processline(char *line)
{
    /* returns 1 for success or 0 for error or 3 for end process */

    int argc=0; /* number of program arguments */
    char **argv=NULL; /* program arguments */
    /* int arg_for=0; */
    pid_t process_id=0;
    int child_process_status=0;

    /* begin */
    process_id=fork();
    if (process_id==(pid_t)0)
    {
        /* child */
        if (line_to_arg(line, &argc, &argv)==0)
        {
            arg_free(&argc, &argv); /* free up memory */
            return 0;
        }

        /*
        for (arg_for=0; arg_for<argc; arg_for++)
            printf("Parameter: >%s<\n", argv[arg_for]); */

        if (argv[0]==NULL) /* if user just pressed return */
        {
            arg_free(&argc, &argv); /* free up memory */
            return 3; /* end child process */
        }
        else
        {
            execvp(argv[0], &argv[0]); /* don't do when user pressed just return or invalid read of size 1 (reported by valgrind) */

            fprintf(stderr, "mysh: error executing: %s\n", argv[0]);

            arg_free(&argc, &argv); /* free up memory */
            return 3; /* end child process */
        }
    }
    else if (process_id!=(pid_t)(-1))
    {
        /* parent */
        wait(&child_process_status);

        line_replace_newline(&line);

        fprintf(stderr, "Exitstatus [%s] = %d\n", line, child_process_status);

        arg_free(&argc, &argv); /* free up memory */
        return 1;
    }
    else
    {
        /* error */
        arg_free(&argc, &argv); /* free up memory */
        return 0;
    }
}

void arg_free(int *argc, char ***argv)
{
    /* frees all memory in passed 'argument array' */

    int arg_for=0;

    if ((*argv)!=NULL)
    {
        for (arg_for=0; arg_for<(*argc); arg_for++)
        {
            if ((*argv)[arg_for]!=NULL) /* don't free non-existing memory of null pointer(s) */
            {
                free((*argv)[arg_for]); /* free data pointers are pointing on */
                (*argv)[arg_for]=NULL;
            }
        }
        free((*argv)); /* free array of pointers */
    }

    (*argc)=0;
}

int line_to_arg(const char *line, int *argc, char ***argv)
{
    /* creates an 'argument array' out of a command line; returns 1 for success or 0 for error */

    char *arg=NULL;
    int arg_start_pos=0;
    int arg_end_pos=0;
    int line_length=0;
    int line_for=0;

    /* NOTE: arguments are bordered by space chars. */

    line_length=strlen(line);

    arg_start_pos=0; /* preset */
    (*argc)=0; /* preset */

    /* create argument array out of inputted line */
    for (line_for=0; line_for<line_length; line_for++)
    {
        if ((line[line_for]==' ')||(line[line_for]=='\n')||(line_for==(line_length-1)))
        {
            if ((line[line_for]==' ')||(line[line_for]=='\n'))
                arg_end_pos=line_for-1;
            else
            {
                fprintf(stderr, "warning in line_to_arg(): no line end sign !");
                arg_end_pos=line_for; /* end of line reached */
            }

            if (arg_end_pos>=arg_start_pos)
            {
                arg=(char *)malloc((arg_end_pos-arg_start_pos+1+1)*sizeof(char)); /* 1 char for null termination */
                if (arg==NULL) return 0;

                memset(arg, 0, (arg_end_pos-arg_start_pos+1+1));
                strncpy(arg, (line+arg_start_pos), (arg_end_pos-arg_start_pos+1));

                (*argc)++;
                (*argv)=(char **)realloc((void *)*argv, (*argc)*sizeof(char **));
                if (argv==NULL) return 0;
                (*argv)[(*argc)-1]=arg;
            }
            arg_start_pos=line_for+1;
        }
    }

    /* add a null pointer required by execvp() */
    (*argc)++;
    (*argv)=(char **)realloc((void *)*argv, (*argc)*sizeof(char **));
    if (argv==NULL) return 0;
    (*argv)[(*argc)-1]=NULL;

    return 1;
}

void line_replace_newline(char **line)
{
    /* replaces \n through \0 */

    int line_length=0;
    int line_for=0;

    line_length=strlen(*line);
    if (line_length>0)
    {
        for (line_for=0; line_for<line_length; line_for++)
        {
            if ((*line)[line_for]=='\n')
            {
                (*line)[line_for]='\0';
                break;
            }
        }
    }
}
