/*
 * Abstract Data Type Joblist
 * 1999/2000 Michael Golm
 * #include <string.h> added 2003 by Matthias Mueller simmmuel Diplominformatik
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "joblist.h"

struct jobinfo
{
    int pid; /* pid of job */
    char *info; /* info describing job */
    struct jobinfo *next; /* next job */
};

struct joblists
{
    struct jobinfo *jobs; /* first element of joblist or NULL */
    struct jobinfo *position; /* iterator position in joblist */
    struct joblists *next; /* next joblist */
};

static struct joblists *lists = NULL;

jl_t jl_new()
{
    struct joblists *l = (struct joblists*) malloc(sizeof(struct joblists));
    l->jobs = NULL;
    l->position = NULL;
    l->next = lists;
    lists = l;
    return (jl_t)l;
}

int jl_delete(jl_t jobs)
{
    struct joblists **l, *t;
    struct jobinfo *j, *c;
    if (jobs == 0) return -1;
    for(l = &lists; *l != NULL; l = &(*l)->next)
    {
        if (*l == (struct joblists*)jobs)
        {
            t = *l;
            *l = (*l)->next;
            /* free all jobs in this joblist */
            for(j = t->jobs; j != NULL;)
            {
                c = j;
                j = j->next;
                free(c->info);
                free(c);
            }
            /* free the joblist */
            free(t);
            return 0;
        }
    }
    return -1;
}

int jl_insert(jl_t jobs, int pid, char* info)
{
    struct jobinfo *job;
    if (jobs == NULL)
    {
        return -1;
    }
    job = (struct jobinfo*) malloc(sizeof(struct jobinfo));
    if (job == NULL)
    {
        return -1;
    }
    job->pid = pid;

    job->info = (char*)malloc((strlen(info)+1) * sizeof(char));
    if (job->info == NULL)
    {
        return -1;
    }
    strcpy(job->info, info);

    job->next = ((struct joblists*)jobs)->jobs;
    ((struct joblists*)jobs)->jobs = job;
    return 0;
}

int jl_remove(jl_t jobs)
{
    struct jobinfo *current, *pos;
    struct jobinfo **j;

    if (jobs == NULL) return -1;

    current = ((struct joblists*)jobs)->position;
    if (current == NULL) return -1;

    pos = NULL;
    for(j = &(((struct joblists*)jobs)->jobs); *j != NULL; j = &(*j)->next)
    {
        if (*j == current)
        {
            *j = (*j)->next;
            free(current);
            ((struct joblists*)jobs)->position = pos;
            return 0;
        }
        pos = *j;
    }

    return -1;
}

int jl_rewind(jl_t jobs)
{
    if (jobs == NULL) return -1;
    ((struct joblists*)jobs)->position = NULL;
    return 0;
}

int jl_next(jl_t jobs, int *pid, char **info)
{
    struct jobinfo *current;

    if (jobs == NULL) return -1;

    current = ((struct joblists*)jobs)->position;

    if (current == NULL)
    {
        /* first jl_next after a call to jl_rewind */
        current = ((struct joblists*)jobs)->jobs;
    }
    else
    {
        current = current->next;
    }
    if (current == NULL)
    {
        /* passed the last element in list */
        return -1;
    }

    ((struct joblists*)jobs)->position = current;

    *pid = current->pid;
    *info = current->info;

    return 0;
}
