#ifndef JOBLIST_H
#define JOBLIST_H

/*
 * A joblist has the type jl_t.
 *
 * To create a new joblist, the jl_new() function
 * must be invoked. When the joblist is no longer
 * needed jl_delete() should be called.
 *
 * To insert an element into the joblist, jl_insert()
 * is used. To remove an element, jl_remove() is used.
 *
 * To iterate through all elements of a joblist,
 * the functions jl_rewind() and jl_next() are used.
 * jl_rewind() rewinds the position to the start of the
 * list. jl_next() sets the position to the next element
 * of the list and writes information about the current
 * element into the integer pid points to and into the
 * char pointer info points to.
 *
 * Example:
 *   char *cmdLine;
 *   int pid;
 *
 *   for(jl_rewind(joblist); jl_next(joblist, &pid, &cmdLine) != -1; ) {
 *      ...
 *   }
 *
 */

typedef struct
{
    int dummy;
} *jl_t;

/*
 * Creates a new joblist.
 *    Upon  successful completion a new joblist
 *    is returned.
 *    Otherwise, NULL is returned.
 */

jl_t jl_new(void);

/*
 * Deletes this joblist.
 *    Upon  successful completion 0 is returned.
 *    Otherwise, -1 is returned.
 */
int jl_delete(jl_t jobs);

/*
 * Inserts an element into the joblist.
 * The string info points to is copied into the joblist. This
 * means that you can free this string after calling jl_insert.
 *    Upon  successful completion 0 is returned.
 *    Otherwise, -1 is returned.
 */
int jl_insert(jl_t jobs, int pid, char *info);

/*
 * Removes the current element from the joblist.
 * The current element is the element for which the
 * most recently called jl_next() returned information.
 *    Upon  successful completion 0 is returned.
 *    Otherwise, -1 is returned.
 */
int jl_remove(jl_t jobs);

/*
 * Sets the position to the start of the joblist.
 *    Upon  successful completion 0 is returned.
 *    Otherwise, -1 is returned.
 */
int jl_rewind(jl_t jobs);

/*
 * Writes information about the current element
 * into pid and info and sets the position to
 * the next element.
 * info must be a pointer to the datastructure pointer.
 *    Upon  successful completion 0 is returned.
 *    Otherwise, -1 is returned.
 */
int jl_next(jl_t jobs, int *pid, char **info);

#endif
