/* Mueller Matthias simmmuel Diplominformatik */
/* Solution (probably) of exercise 'timed' */
/* 19.12.03 - 06.01.04 */

/* includes */
#include <stdlib.h>     /* exit */
#include <stdio.h>      /* printf */
#include <errno.h>      /* perror */
#include <unistd.h>     /* getuid */
#include <sys/types.h>  /* getuid */
#include <sys/socket.h> /* socket stuff */
#include <netinet/in.h> /* socketadd_in structure */
#include <time.h>       /* time */
#include <sys/wait.h>   /* wait */
#include <signal.h>     /* [signal handler] */

/* prototypes */
int main();
int socket_create_connection(int socket_descr, int *socket_descr_accepted, int port_number);
void free_memory(int *socket_descr_accepted, struct tm **tm_var, char **buffer, time_t **current_time);
void signal_handler(int signal);

int main()
{
    /* connection creation */
    int socket_descr=0; /* socket descriptor */
    int socket_descr_accepted=0; /* socket descriptor (accepted socket) */
    struct sockaddr_in *socketaddr_in_var=NULL;
    int port_number=0; /* port number used for connection (must be >= 1024) */
    /* time string creation */
    /* int buffer_size=0; */
    char *buffer=NULL; /* what is sent to client */
    time_t *current_time=NULL; /* current time (pointer to a time_t-var) */
    struct tm *tm_var=NULL;
    pid_t process_id=0;
    /* other */
    struct sigaction sigaction_var;

    /* reset */
    errno=0;

    /* preset */
    port_number=(int)getuid(); /* no error possible (they say ;-P) */
    if (port_number<1024) /* error handling (necessary on my home machine (SuSE 8.1, logged in as root user)) */
    {
        fprintf(stderr, "Can't use UID (%d) as port number, now using: 1024\n", port_number);
        port_number=1024;
    }

    buffer=(char *)malloc((size_t)128); /* allocate some more space as required to avoid errors */
    if (buffer==NULL)
    {
        perror("Error allocating memory");
        free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
        exit(EXIT_FAILURE);
    }

    current_time=(time_t *)malloc(sizeof(time_t));
    if (current_time==NULL)
    {
        perror("Error allocating memory");
        free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
        exit(EXIT_FAILURE);
    }

    /* begin */

    /* intiialize socket connection */
    socket_descr = socket(PF_INET, SOCK_STREAM, 0);
    if (socket_descr==(-1))
    {
        fprintf(stderr, "Error opening socket connection in main !");
        free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
        exit(EXIT_FAILURE);
    }

    /* bind socket connection to port #port_number */
    socketaddr_in_var=(struct sockaddr_in *)malloc((size_t)sizeof(struct sockaddr_in));
    if (socketaddr_in_var==NULL)
    {
        fprintf(stderr, "Error allocating memory for var socketaddr_in_var in main !");
        free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
        exit(EXIT_FAILURE);
    }

    socketaddr_in_var->sin_family = AF_INET;
    socketaddr_in_var->sin_port = htons(port_number);
    socketaddr_in_var->sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(socket_descr, (const struct sockaddr *)socketaddr_in_var, sizeof(*socketaddr_in_var))==(-1))
    {
        perror("Error binding socket connection in main");
        free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
        exit(EXIT_FAILURE);
    }

    /* install signal handler to catch zombie-children */
    sigaction_var.sa_handler = signal_handler;
    sigemptyset(&sigaction_var.sa_mask);
    sigaction_var.sa_flags = SA_RESTART; /* important to use SA_RESTART here, or we'll get an 'interrupted system call' error when a client is terminated */

    errno=0; /* may have been set by any signal handler */
    if (sigaction(SIGCHLD, &sigaction_var, NULL)!=0)
    {
        perror("Error in sigaction() in main") ;
        free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
        exit(EXIT_FAILURE);
    }

    /* enter loop to create clients */
    while (1)
    {

        if (socket_create_connection(socket_descr, &socket_descr_accepted, port_number)==0)
        {
            /* socket_create_connection() already created error message */
            free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
            exit(EXIT_FAILURE);
        }

        /* create a child process that serves as server, and make this process receive further requests */
        process_id=fork();
        if (process_id==(pid_t)0)
        {
            /* child */
            /* send time permanently */
            while (1)
            {
                if (time(current_time)==(time_t)(-1))
                {
                    perror("Error determining time in main");
                    free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
                    exit(EXIT_FAILURE);
                }

                if (tm_var==NULL)
                    tm_var=(struct tm *)malloc(sizeof(struct tm)); /* error checking below */

                if (tm_var==NULL)
                {
                    fprintf(stderr, "Error allocating space for var tm_var in main !");
                    free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
                    exit(EXIT_FAILURE);
                }

                tm_var=localtime(current_time);

                if (tm_var==NULL)
                {
                    fprintf(stderr, "localtime() failed !");
                    free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
                    exit(EXIT_FAILURE);
                }

                strftime(buffer, 18, "%d.%m.%Y %H:%M\n", tm_var);

                if (write(socket_descr_accepted, buffer, 18)==(size_t)(-1))
                {
                    perror("Error sending data buffer"); /* continue, maybe merely high network load */
                    errno=0; /* reset */
                }

                if (sleep(1)>0) /* error-handling */
                    fprintf(stderr, "DAMN!! - sleep(1) failed !!!");
            }
        }
        else if (process_id!=(pid_t)(-1))
        {
            /* parent */
            /* continue in loop, open new socket and wait for connection request */
        }
        else
        {
            /* error */
            fprintf(stderr, "Error creating child process in main !");
            free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
            exit(EXIT_FAILURE);
        }
    }

    /* get out of here */
    free_memory(&socket_descr_accepted, &tm_var, &buffer, &current_time);
    return 0;
}

int socket_create_connection(int socket_descr, int *socket_descr_accepted, int port_number)
{
    /* opens a socket connection for passed port number, listens and returns 1 (descriptor for accepted socket connection then set) for success or 0 for error */

    /* open port and listen */
    if (listen(socket_descr, 5)==(-1))
    {
        perror("Error listening for incoming connection requests in socket_create_connection()");
        return 0;
    }

    /* accept connection */
    *socket_descr_accepted=accept(socket_descr, NULL, 0);
    if (*socket_descr_accepted==(-1))
    {
        perror("Error accepting incoming connection request in socket_create_connection()");
        return 0;
    }

    return 1;
}

void free_memory(int *socket_descr_accepted, struct tm **tm_var, char **buffer, time_t **current_time)
{
    /* free up memory and close open files/sockets */

    if (*socket_descr_accepted)
    {
        close(*socket_descr_accepted);
        *socket_descr_accepted=0;
    }
    if (*tm_var!=NULL)
    {
        free(*tm_var);
        *tm_var=NULL;
    }
    if (*buffer!=NULL)
    {
        free(*buffer);
        *buffer=NULL;
    }
    if (*current_time!=NULL)
    {
        free(*current_time);
        *current_time=NULL;
    }
}

void signal_handler(int signal)
{
    /* receives signals of child processes */

    pid_t wait_pid;
    int wait_status=0;

    switch (signal)
    {
    case SIGCHLD: /* status of a child changed */

        while ((wait_pid = waitpid (-1, &wait_status, WNOHANG)) > 0)
        {
            /* child process resources freed through waitpid()-call */
        }
    }
    return;
}
