/* Mueller Matthias simmmuel Diplominformatik */
/* Solution (probably) of exercise 'port_forward' */
/* 12.1.04 - 15.01.04 */

/* NOTE: this project was built out of a previous exercise, 'timed'. */

/* includes */
#include <stdlib.h>     /* exit */
#include <stdio.h>      /* printf */
#include <errno.h>      /* perror */
#include <unistd.h>     /* getuid */
#include <sys/types.h>  /* getuid */
#include <sys/socket.h> /* socket stuff */
#include <netinet/in.h> /* socketadd_in structure */
#include <netdb.h>      /* gethostbyname */
#include <sys/wait.h>   /* wait, waitpid */
#include <string.h>     /* memset, memcpy */
#include <signal.h>     /* [signal handler] */
#include "forward.h"    /* forward-function */

/* defines */
#define h_addr  h_addr_list[0] /* see H-Uebung6-A5.pdf */

/* prototypes */
int main(int argc, char *argv[]);
int socket_create_connection(int socket_descr, int *socket_descr_accepted, int port_number);
void free_memory(int *socket_descr_accepted, struct sockaddr_in **socketaddr_in_var);
void free_forward_memory(int *forward_socket_descr, struct hostent **hostent_var, struct sockaddr_in **forward_socket_addr_in_var);
void signal_handler(int signal);

int main(int argc, char *argv[])
{
    /* local connection creation */
    int port_number=0; /* port number used for client-port_forward connection (must be >= 1024) */
    int socket_descr=0; /* socket descriptor */
    int socket_descr_accepted=0; /* socket descriptor (accepted socket) */
    struct sockaddr_in *socketaddr_in_var=NULL;

    /* forking */
    pid_t process_id=0;

    /* forward connection creation */
    int forward_port_number=0; /* port number used for port_forward-server connection (must be >= 1024) */
    int forward_socket_descr=0;
    struct hostent *hostent_var=NULL;
    struct sockaddr_in *forward_socketaddr_in_var=NULL;

    /* other */
    struct sigaction sigaction_var;
    int temp=0;

    /* reset */
    errno=0;

    /* verify */
    if (argc<3)
    {
        fprintf(stderr, "usage: pforward <local port> <forward host> <forward port>\n");
        free_memory(&socket_descr_accepted, &socketaddr_in_var);
        free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
        exit(EXIT_FAILURE);
    }

    /* preset */

    /* display arguments */
    for (temp=0; temp<argc; temp++)
        printf("%s ", argv[temp]);
    printf("\n");

    /* process passed port numbers */
    port_number=atoi(argv[1]); /* local port */
    if ((port_number<1024)||(port_number>65535)) /* verify */
    {
        fprintf(stderr, "Can't use passed local port number (%d), now using: 1024\n", port_number);
        port_number=1024;
    }

    forward_port_number=atoi(argv[3]); /* (int)getuid(); */
    if ((forward_port_number<0)||(forward_port_number>65535)) /* verify */
    {
        fprintf(stderr, "Can't use passed forward port number (%d), now using: 80\n", forward_port_number);
        forward_port_number=80;
    }

    /* begin */

    /* intiialize socket connection */
    socket_descr = socket(PF_INET, SOCK_STREAM, 0);
    if (socket_descr==(-1))
    {
        fprintf(stderr, "Error opening socket connection in main !");
        free_memory(&socket_descr_accepted, &socketaddr_in_var);
        free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
        exit(EXIT_FAILURE);
    }
    else
    {
        printf("created server socket!\n");
    }

    /* bind socket connection to port #port_number */
    socketaddr_in_var=(struct sockaddr_in *)malloc((size_t)sizeof(struct sockaddr_in));
    if (socketaddr_in_var==NULL)
    {
        fprintf(stderr, "Error allocating memory for var socketaddr_in_var in main !");
        free_memory(&socket_descr_accepted, &socketaddr_in_var);
        free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
        exit(EXIT_FAILURE);
    }

    socketaddr_in_var->sin_family=AF_INET;
    socketaddr_in_var->sin_port=htons(port_number);
    socketaddr_in_var->sin_addr.s_addr=htonl(INADDR_ANY);
    if (bind(socket_descr, (const struct sockaddr *)socketaddr_in_var, sizeof(*socketaddr_in_var))==(-1))
    {
        perror("Error binding socket connection in main()");
        free_memory(&socket_descr_accepted, &socketaddr_in_var);
        free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
        exit(EXIT_FAILURE);
    }
    else
    {
        printf("bound server socket to port %d!\n", port_number);
    }

    /* install signal handler to catch zombie-children */
    sigaction_var.sa_handler = signal_handler;
    sigemptyset(&sigaction_var.sa_mask);
    sigaction_var.sa_flags = SA_RESTART; /* important to use SA_RESTART here, or we'll get an 'interrupted system call' error when a client is terminated */

    errno=0; /* may have been set by any signal handler */
    if (sigaction(SIGCHLD, &sigaction_var, NULL)!=0)
    {
        perror("Error in sigaction() in main()") ;
        free_memory(&socket_descr_accepted, &socketaddr_in_var);
        free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
        exit(EXIT_FAILURE);
    }

    /* enter loop to create clients */
    while (1)
    {
        printf("server socket is now listening!\n");
        if (socket_create_connection(socket_descr, &socket_descr_accepted, port_number)==0)
        {
            /* socket_create_connection() already created error message */
            free_memory(&socket_descr_accepted, &socketaddr_in_var);
            free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
            exit(EXIT_FAILURE);
        }
        else
            printf("request!\n");

        /* create a child process that serves as server, and make this process receive further requests */
        process_id=fork();
        if (process_id==(pid_t)0)
        {
            /* child */

            /* get address of target host */
            forward_socket_descr = socket(PF_INET, SOCK_STREAM, 0);
            if (forward_socket_descr==(-1))
            {
                fprintf(stderr, "Error opening forward socket connection in main !");
                free_memory(&socket_descr_accepted, &socketaddr_in_var);
                free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
                exit(EXIT_FAILURE);
            }
            else
            {
                printf("created client socket!\n");
            }

            hostent_var=(struct hostent *)malloc(sizeof(struct hostent));
            if (hostent_var==NULL)
            {
                perror("Error allocating memory for hostent_var in main()");
                free_memory(&socket_descr_accepted, &socketaddr_in_var);
                free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
                exit(EXIT_FAILURE);
            }

            hostent_var=gethostbyname(argv[2]);
            if (hostent_var==NULL)
            {
                perror("Error in gethostbyname() in main()");
                free_memory(&socket_descr_accepted, &socketaddr_in_var);
                free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
                exit(EXIT_FAILURE);
            }
            else
                printf("got address of %s\n", argv[2]); /* -> %d.%d.%d.%d\n", argv[2],
                    (int)*((hostent_var->h_addr)+0),
                    (int)*((hostent_var->h_addr)+1),
                    (int)*((hostent_var->h_addr)+2),
                    (int)*((hostent_var->h_addr)+3)); */ /* rest does not work :-( */

            forward_socketaddr_in_var=(struct sockaddr_in *)malloc((size_t)sizeof(struct sockaddr_in));
            if (forward_socketaddr_in_var==NULL)
            {
                fprintf(stderr, "Error allocating memory for var forward_socketaddr_in_var in main() !");
                free_memory(&socket_descr_accepted, &socketaddr_in_var);
                free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
                exit(EXIT_FAILURE);
            }

            memset(forward_socketaddr_in_var, 0, sizeof(*forward_socketaddr_in_var)); /* initialize structure */
            memcpy((char *)(&(forward_socketaddr_in_var->sin_addr)), (char *)hostent_var->h_addr, hostent_var->h_length);
            forward_socketaddr_in_var->sin_family = AF_INET;
            forward_socketaddr_in_var->sin_port = htons(forward_port_number);

            printf("start accepting connections!\n");
            if (connect(forward_socket_descr, (struct sockaddr *)forward_socketaddr_in_var, sizeof(*forward_socketaddr_in_var))==(-1))
            {
                perror("Error connecting to forward host");
                errno=0;
            }
            else
            {
                printf("accept!");
            }

            /* prinf("connected!\naccepted!\starting service\nwhatever\n"); */ /* hat ja keiner gesagt, dass wir das machen sollen, und schon gleich gar nicht, wie */

            /* forward data using a special forward-function (forward.c) */
            if (forward(socket_descr_accepted, forward_socket_descr)==(-1))
            {
                perror("Error forwarding data");
                errno=0;
            }

            /* close(socket_descr_accepted); */ /* CLOSE 2/2 */
            free_memory(&socket_descr_accepted, &socketaddr_in_var); /* CLOSE 2/2 */
            free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);

            return 0; /* get out of here, or we'll get into serious troubles after 15 minutes Konqueror surfing */
        }
        else if (process_id!=(pid_t)(-1))
        {
            /* parent */
            close(socket_descr_accepted); /* CLOSE 1/2; close one of two file descriptors (the one of parent and later the one of child) to verify free_memory()-call of child will finally close file object */
            /* continue in loop, open new socket and wait for connection request */
        }
        else
        {
            /* error */
            fprintf(stderr, "Error creating child process in main !");
            free_memory(&socket_descr_accepted, &socketaddr_in_var);
            free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
            exit(EXIT_FAILURE);
        }
    }

    /* get out of here */
    free_memory(&socket_descr_accepted, &socketaddr_in_var);
    free_forward_memory(&forward_socket_descr, &hostent_var, &forward_socketaddr_in_var);
    return 0;
}

int socket_create_connection(int socket_descr, int *socket_descr_accepted, int port_number)
{
    /* opens a socket connection for passed port number, listens and returns 1 (descriptor for accepted socket connection then set) for success or 0 for error */

    /* open port and listen */
    if (listen(socket_descr, 5)==(-1))
    {
        perror("Error listening for incoming connection requests in socket_create_connection()");
        return 0;
    }

    /* accept connection */
    *socket_descr_accepted=accept(socket_descr, NULL, 0);
    if (*socket_descr_accepted==(-1))
    {
        perror("Error accepting incoming connection request in socket_create_connection()");
        return 0;
    }

    return 1;
}

void free_memory(int *socket_descr_accepted, struct sockaddr_in **socketaddr_in_var)
{
    /* free up memory and close open files/sockets */

    if (*socket_descr_accepted)
    {
        close(*socket_descr_accepted); /* ATTENTION: child AND parent must close socket! */
        *socket_descr_accepted=0;
    }
    if (*socketaddr_in_var!=NULL)
    {
        free(*socketaddr_in_var);
        *socketaddr_in_var=NULL;
    }
}

void free_forward_memory(int *forward_socket_descr, struct hostent **hostent_var, struct sockaddr_in **forward_socket_addr_in_var)
{
    /* free up memory and close open files/sockets of 'forward section' */
    if (*forward_socket_descr)
    {
        close(*forward_socket_descr);
        *forward_socket_descr=0;
    }
    if ((*hostent_var)!=NULL)
    {
        free(*hostent_var);
        *hostent_var=NULL;
    }
    if ((*forward_socket_addr_in_var)!=NULL)
    {
        free(*forward_socket_addr_in_var);
        *forward_socket_addr_in_var=NULL;
    }
}

void signal_handler(int signal)
{
    /* receives signals of child processes */

    pid_t wait_pid;
    int wait_status=0;

    switch (signal)
    {
    case SIGCHLD: /* status of a child changed */

        while ((wait_pid=waitpid(-1, &wait_status, WNOHANG))>0)
        {
            /* child process resources freed through waitpid()-call */
        }
    }
    return;
}
